class TrailwatchError(Exception):
    """Base exception for all trailwatch exceptions"""


class NotConfiguredError(TrailwatchError):
    """Exception raised when trailwatch is not configured"""


class ExecutionTimeoutError(TrailwatchError):
    """Exception raised when an execution times out"""


class PartialSuccessError(TrailwatchError):
    """Exception raised when an execution was partially successful"""



import datetime

from types import TracebackType
from typing import Type

from .config import DEFAULT, Default, TrailwatchConfig
from .connectors.base import Connector
from .exceptions import ExecutionTimeoutError, PartialSuccessError, TrailwatchError


class TrailwatchContext:
    def __init__(
        self,
        job: str,
        job_description: str,
        loggers: list[str] | Default | None = DEFAULT,
        execution_ttl: int | Default | None = DEFAULT,
        log_ttl: int | Default | None = DEFAULT,
        error_ttl: int | Default | None = DEFAULT,
    ) -> None:
        """
        Initialize a TrailwatchContext instance for a job.

        Parameters
        ----------
        job : str
            Job name. E.g., 'Upsert appointments'.
        job_description : str
            Job description. E.g., 'Upsert appointments from ModMed to Salesforce'.
        loggers : list[str], optional
            List of loggers logs from which are sent to TrailWatch.
            By default, no logs are sent.
        execution_ttl : int, optional
            Time to live for the execution record in seconds.
        log_ttl : int, optional
            Time to live for the log records in seconds.
        error_ttl : int, optional
            Time to live for the error records in seconds.

        """
        self.config = TrailwatchConfig(
            job=job,
            job_description=job_description,
            loggers=loggers,
            execution_ttl=execution_ttl,
            log_ttl=log_ttl,
            error_ttl=error_ttl,
        )
        self.connectors: list[Connector] = []

    def __enter__(self) -> "TrailwatchContext":
        for connector_factory in self.config.shared_configuration.connectors:
            connector = connector_factory(self.config)
            connector.start_execution()
            self.connectors.append(connector)
        return self

    def __exit__(
        self,
        exc_type: Type[Exception] | None,
        exc_value: Exception | None,
        exc_traceback: TracebackType | None,
    ) -> bool:
        end = datetime.datetime.utcnow()
        if exc_type is None:
            status = "success"
        else:
            if issubclass(exc_type, ExecutionTimeoutError):
                status = "timeout"
            elif issubclass(exc_type, PartialSuccessError):
                status = "partial"
            else:
                status = "failure"
        for connector in self.connectors:
            connector.finalize_execution(status, end)
            if exc_type is not None and not issubclass(exc_type, TrailwatchError):
                assert exc_value is not None
                assert exc_traceback is not None
                connector.handle_exception(
                    timestamp=end,
                    exc_type=exc_type,
                    exc_value=exc_value,
                    exc_traceback=exc_traceback,
                )
        if isinstance(exc_type, PartialSuccessError):
            return True
        return False



import datetime
import logging

from typing import Optional

from requests import Response, Session

logger = logging.getLogger(__name__)


class TrailwatchApi:
    def __init__(self, session: Session, url: str, api_key: str) -> None:
        self.session = session
        self.url = url
        self.api_key = api_key

    def _make_request(self, method: str, url: str, **kwargs) -> Optional[Response]:
        try:
            response = self.session.request(
                method,
                url,
                headers={"x-api-key": self.api_key},
                timeout=30,
                **kwargs,
            )
            response.raise_for_status()
            return response
        except Exception as error:
            logger.error(
                "Failed to make '%s' request to '%s' due to: '%s'",
                method,
                url,
                error,
            )
            return None

    def upsert_project(self, name: str, description: str) -> None:
        self._make_request(
            "PUT",
            "/".join([self.url, "api", "v1", "projects"]),
            json={"name": name, "description": description},
        )

    def upsert_environment(self, name: str) -> None:
        self._make_request(
            "PUT",
            "/".join([self.url, "api", "v1", "environments"]),
            json={"name": name},
        )

    def upsert_job(self, name: str, description: str, project: str) -> None:
        self._make_request(
            "PUT",
            "/".join([self.url, "api", "v1", "jobs"]),
            json={"name": name, "description": description, "project": project},
        )

    def create_execution(
        self,
        project: str,
        environment: str,
        job: str,
        ttl: Optional[int],
    ) -> Optional[str]:
        try:
            response = self._make_request(
                "POST",
                "/".join([self.url, "api", "v1", "executions"]),
                json={
                    "project": project,
                    "environment": environment,
                    "job": job,
                    "status": "running",
                    "start": datetime.datetime.utcnow().isoformat(),
                    "ttl": ttl,
                },
            )
            if response is None:
                return None
            return response.json()["id"]
        except KeyError as error:
            logger.error("Failed to create execution due to: '%s'", error)
            return None

    def create_log(
        self,
        execution_id: str,
        timestamp: datetime.datetime,
        name: str,
        levelno: int,
        lineno: int,
        msg: str,
        func: str,
        ttl: Optional[int],
    ) -> None:
        self._make_request(
            "POST",
            "/".join([self.url, "api", "v1", "logs"]),
            json={
                "execution_id": execution_id,
                "timestamp": timestamp.isoformat(),
                "name": name,
                "levelno": levelno,
                "lineno": lineno,
                "msg": msg,
                "func": func,
                "ttl": ttl,
            },
        )

    def update_execution(
        self,
        execution_id: str,
        status: str,
        end: datetime.datetime,
    ) -> None:
        self._make_request(
            "PATCH",
            "/".join([self.url, "api", "v1", "executions", execution_id]),
            json={"status": status, "end": end.isoformat()},
        )

    def create_error(
        self,
        execution_id: str,
        timestamp: datetime.datetime,
        name: str,
        message: str,
        traceback: str,
        ttl: Optional[int],
    ) -> None:
        self._make_request(
            "POST",
            "/".join([self.url, "api", "v1", "errors"]),
            json={
                "execution_id": execution_id,
                "timestamp": timestamp.isoformat(),
                "name": name,
                "msg": message,
                "ttl": ttl,
                "traceback": traceback,
            },
        )



from dataclasses import dataclass, field

from trailwatch.connectors.base import ConnectorFactory

from .exceptions import NotConfiguredError


class Default:
    """
    Sentinel value to indicate that user did not set a value.

    Used to distinguish between None and using globally configured value.

    """


DEFAULT = Default()


@dataclass(frozen=True)
class SharedConfiguration:
    is_configured: bool = False
    connectors: list[ConnectorFactory] = field(default_factory=list)

    project: str = ""
    project_description: str = ""
    environment: str = ""
    loggers: list[str] = field(default_factory=list)
    execution_ttl: int | None = None
    log_ttl: int | None = None
    error_ttl: int | None = None


class TrailwatchConfig:
    shared_configuration: SharedConfiguration = SharedConfiguration()

    # Job-specific properties
    job: str
    job_description: str

    # Shared properties which can be overridden for each job
    _loggers: list[str] | Default | None
    _execution_ttl: int | Default | None
    _log_ttl: int | Default | None
    _error_ttl: int | Default | None

    def __init__(
        self,
        job: str,
        job_description: str,
        loggers: list[str] | Default | None = DEFAULT,
        execution_ttl: int | Default | None = DEFAULT,
        log_ttl: int | Default | None = DEFAULT,
        error_ttl: int | Default | None = DEFAULT,
    ) -> None:
        """
        Initialize a TrailwatchConfig instance for a job.

        Parameters
        ----------
        job : str
            Job name. E.g., 'Upsert appointments'.
        job_description : str
            Job description. E.g., 'Upsert appointments from ModMed to Salesforce'.
        loggers : list[str], optional
            List of loggers logs from which are sent to TrailWatch.
            By default, no logs are sent.
        execution_ttl : int, optional
            Time to live for the execution record in seconds.
            By default, global configuration is used.
        log_ttl : int, optional
            Time to live for the log records in seconds.
            By default, global configuration is used.
        error_ttl : int, optional
            Time to live for the error records in seconds.
            By default, global configuration is used.

        """
        if not self.shared_configuration.is_configured:
            raise NotConfiguredError(
                "TrailWatch must be configured with trailwatch.configure() before using"
            )
        self.job = job
        self.job_description = job_description
        self._loggers = loggers or []
        self._execution_ttl = execution_ttl
        self._log_ttl = log_ttl
        self._error_ttl = error_ttl

    @property
    def project(self) -> str:
        return self.shared_configuration.project

    @property
    def project_description(self) -> str:
        return self.shared_configuration.project_description

    @property
    def environment(self) -> str:
        return self.shared_configuration.environment

    @property
    def loggers(self) -> list[str]:
        if self._loggers is not DEFAULT:
            assert isinstance(self._loggers, (list, type(None)))
            return self._loggers or []
        return self.shared_configuration.loggers

    @property
    def execution_ttl(self) -> int | None:
        if self._execution_ttl is not DEFAULT:
            assert isinstance(self._execution_ttl, (int, type(None)))
            return self._execution_ttl
        return self.shared_configuration.execution_ttl

    @property
    def log_ttl(self) -> int | None:
        if self._log_ttl is not DEFAULT:
            assert isinstance(self._log_ttl, (int, type(None)))
            return self._log_ttl
        return self.shared_configuration.log_ttl

    @property
    def error_ttl(self) -> int | None:
        if self._error_ttl is not DEFAULT:
            assert isinstance(self._error_ttl, (int, type(None)))
            return self._error_ttl
        return self.shared_configuration.error_ttl


def configure(
    project: str,
    project_description: str,
    environment: str,
    connectors: list[ConnectorFactory],
    loggers: list[str] | None = None,
    execution_ttl: int | None = None,
    log_ttl: int | None = None,
    error_ttl: int | None = None,
):
    """
    Configure TrailWatch.

    This function configures global settings shared across all executions unless
    explicitly overridden for a specific job.
    This function must be called once per process.

    Parameters
    ----------
    project : str
        Short name of the project.
        E.g., 'companyname-middleware'.
    project_description : str
        User-friendly description of the project.
        E.g., 'Synchronize appointment data from ModMed with companyname's Salesforce'.
    environment : str
        Environment in which all jobs are executed.
        E.g., 'production', 'staging', 'development'.
    connectors : list[ConnectorFactoryBase]
        List of connectors to use.
    loggers : list[str], optional
        List of loggers logs from which are sent to TrailWatch.
        By default, no logs are sent.
    execution_ttl : int, optional
        Time to live for the execution record in seconds.
        By default, the execution record is kept indefinitely.
    log_ttl : int, optional
        Time to live for the log records in seconds.
        By default, the log records are kept indefinitely.
    error_ttl : int, optional
        Time to live for the error records in seconds.
        By default, the error records are kept indefinitely.

    """
    if len(connectors) == 0:
        raise ValueError("At least one connector must be configured")
    TrailwatchConfig.shared_configuration = SharedConfiguration(
        is_configured=True,
        project=project,
        project_description=project_description,
        environment=environment,
        loggers=loggers or [],
        execution_ttl=execution_ttl,
        log_ttl=log_ttl,
        error_ttl=error_ttl,
        connectors=connectors,
    )



__all__ = [
    "configure",
    "TrailwatchContext",
    "watch",
]

import functools
import signal

from .config import DEFAULT, Default, configure
from .context import TrailwatchContext
from .exceptions import ExecutionTimeoutError


def throw_timeout_on_alarm(signum, frame):
    raise ExecutionTimeoutError


signal.signal(signal.SIGALRM, throw_timeout_on_alarm)


# TODO - add support for uploading files (use locals() to get execution context)
def watch(
    job: str | None = None,
    job_description: str | None = None,
    loggers: list[str] | Default | None = DEFAULT,
    execution_ttl: int | Default | None = DEFAULT,
    log_ttl: int | Default | None = DEFAULT,
    error_ttl: int | Default | None = DEFAULT,
    timeout: int | None = None,
):
    """
    Watch a callable (function or method).

    This is a decorator that can be used to wrap a callable (function or method)
    and send execution statistics (start, end, name, logs, exceptions, etc.)
    to configured connectors.

    Parameters
    ----------
    job : str
        Job name. E.g., 'Upsert appointments'.
    job_description : str
        Job description. E.g., 'Upsert appointments from ModMed to Salesforce'.
    loggers : list[str], optional
        List of loggers logs from which are sent to TrailWatch.
        By default, no logs are sent.
    execution_ttl : int, optional
        Time to live for the execution record in seconds.
        By default, global configuration is used.
    log_ttl : int, optional
        Time to live for the log records in seconds.
        By default, global configuration is used.
    error_ttl : int, optional
        Time to live for the error records in seconds.
        By default, global configuration is used.
    timeout : int, optional
        Timeout in seconds. If the callable takes longer than this to execute,
        an execution timeout error is raised and execution is marked as timed out.
        By default, no timeout is set.

    """

    def wrapper(func):
        decorator_kwargs = {
            "job": job or func.__name__,
            "job_description": job_description or func.__doc__,
            "loggers": loggers,
            "execution_ttl": execution_ttl,
            "log_ttl": log_ttl,
            "error_ttl": error_ttl,
        }
        if decorator_kwargs["job_description"] is None:
            raise ValueError(
                "Job description must either be provided explicitly or "
                "via the docstring of the decorated function"
            )

        @functools.wraps(func)
        def inner(*args, **kwargs):
            if timeout is not None:
                signal.alarm(timeout)
            with TrailwatchContext(**decorator_kwargs):
                return func(*args, **kwargs)

        return inner

    return wrapper



import datetime

from abc import ABC, abstractmethod
from types import TracebackType
from typing import TYPE_CHECKING, Type

if TYPE_CHECKING:
    from trailwatch.config import TrailwatchConfig


class Connector(ABC):
    @abstractmethod
    def start_execution(self) -> None:
        """Create execution record, attach handlers, etc."""

    @abstractmethod
    def finalize_execution(self, status: str, end: datetime.datetime) -> None:
        """Update execution record, detach handlers, etc."""

    @abstractmethod
    def handle_exception(
        self,
        timestamp: datetime.datetime,
        exc_type: Type[Exception],
        exc_value: Exception,
        exc_traceback: TracebackType,
    ):
        """Add exception to execution record (or do nothing)."""


class ConnectorFactory(ABC):
    @abstractmethod
    def __call__(self, config: "TrailwatchConfig") -> Connector:
        """Create connector instance."""



import datetime
import logging

from typing import Optional

from .api import TrailwatchApi


class AwsHandler(logging.Handler):
    def __init__(
        self,
        execution_id: str,
        api: TrailwatchApi,
        ttl: Optional[int] = None,
    ):
        logging.Handler.__init__(self)
        self.execution_id = execution_id
        self.api = api
        self.ttl = ttl

    def emit(self, record: logging.LogRecord):
        self.format(record)
        self.api.create_log(
            execution_id=self.execution_id,
            timestamp=datetime.datetime.utcfromtimestamp(record.created),
            name=record.name,
            levelno=record.levelno,
            lineno=record.lineno,
            msg=record.message,
            func=record.funcName,
            ttl=self.ttl,
        )



import datetime
import logging
import traceback

from types import TracebackType
from typing import TYPE_CHECKING, Type

from requests import Session

from trailwatch.connectors.base import Connector, ConnectorFactory

from .api import TrailwatchApi
from .handler import AwsHandler

if TYPE_CHECKING:
    from trailwatch.config import TrailwatchConfig

logger = logging.getLogger(__name__)


class AwsConnector(Connector):
    def __init__(self, config: "TrailwatchConfig", url: str, api_key: str) -> None:
        self.config = config
        self.api = TrailwatchApi(Session(), url, api_key)
        self.execution_id: str | None = None
        self.handler: AwsHandler | None = None

    def start_execution(self) -> None:
        # Create entries in TrailWatch database
        self.api.upsert_project(
            self.config.project,
            self.config.project_description,
        )
        self.api.upsert_environment(self.config.environment)
        self.api.upsert_job(
            self.config.job,
            self.config.job_description,
            self.config.project,
        )
        self.execution_id = self.api.create_execution(
            self.config.project,
            self.config.environment,
            self.config.job,
            self.config.execution_ttl,
        )

        # Register logging handlers
        if self.execution_id is not None:
            self.handler = AwsHandler(
                self.execution_id,
                self.api,
                self.config.log_ttl,
            )
            for logger_name in self.config.loggers:
                logging.getLogger(logger_name).addHandler(self.handler)

    def finalize_execution(self, status: str, end: datetime.datetime) -> None:
        if self.execution_id is not None:
            self.api.update_execution(self.execution_id, status, end)

        # Remove logging handlers
        if self.handler is not None:
            for logger_name in self.config.loggers:
                logging.getLogger(logger_name).removeHandler(self.handler)

    def handle_exception(
        self,
        timestamp: datetime.datetime,
        exc_type: Type[Exception],
        exc_value: Exception,
        exc_traceback: TracebackType,
    ):
        if self.execution_id is not None:
            self.api.create_error(
                execution_id=self.execution_id,
                timestamp=timestamp,
                name=exc_type.__name__,
                message=str(exc_value),
                traceback="".join(
                    traceback.format_exception(
                        exc_type,
                        value=exc_value,
                        tb=exc_traceback,
                    )
                ),
                ttl=self.config.error_ttl,
            )


class AwsConnectorFactory(ConnectorFactory):
    def __init__(self, url: str, api_key: str) -> None:
        """
        Initialize TrailWatch AWS connector factory.

        Parameters
        ----------
        url : str
            URL pointing to TrailWatch instance deployed on AWS.
            E.g., 'https://somerandomstring.execute-api.us-west-2.amazonaws.com'.
        api_key : str
            API key to be included in the 'x-api-key' header when calling the REST API.

        """
        self.url = url
        self.api_key = api_key

    def __call__(self, config: "TrailwatchConfig") -> AwsConnector:
        return AwsConnector(config, self.url, self.api_key)
